/**************************************************************************************************************************
// This script was authored by Lorne Brooks - Email: lbtools2k@gmail.com
//
// It was inspired by this reddit posting... https://www.reddit.com/r/Maxscript/comments/20h97g/script_evenly_spread_out_objects_on_xyz/
// I modified the behaviour of the spread algorithm slighly, but it's basically the same - so full credit to the author in that reddit posting.
//
//
// REVISION HISTORY:
// Version 1.1 15/11/18;
//						- Added absolute and relative options to offset by value. Relative by default. Absolute mode will ignore zero values.
//						- All options now correctly find selected objects when the UI is opened without an initial selection.
// Version 1.0 09/11/18;
//						- First public release.
//
//
// This script is not copyrighted, please feel free to modify or distribute.
//
**************************************************************************************************************************/

global g_LBTools_spreadObjs

struct s_LBTools_spreadObjs
(
/*==================================================================================
// Structure variables
====================================================================================*/

spreadObjects_Rollout,
aboutScript_rollout,
spreadObjectsINI = getDir #plugcfg + @"\LBTools_SpreadObjects.ini",
dPos,
	
/*==================================================================================*/

fn f_spreadObjects =
(
	rollout spreadObjects_Rollout "Spread Objects" width:130 height:96
	(
		local spreadX_val = 0.0
		local spreadY_val = 0.0
		local spreadZ_val = 0.0
		local offsetX_val = 0.0
		local offsetY_val = 0.0
		local offsetZ_val = 0.0
		local a_selObjs = #()
		local a_posX = #()	
		local a_posY = #()
		local a_posZ = #()
		local useSelOrder = false
		local doReverseOrder = false
		local doByNameOrder = true
		local doByDistanceOrder = false
		local doByVolumeOrder = false
		
		--// Finds the volume and centre of mass - returns a list [Vol, COM]
		--// Taken from MXS help docs ... https://help.autodesk.com/view/3DSMAX/2017/ENU/?guid=__files_GUID_B94A1094_002C_4A7D_B5DD_AF5AB3435041_htm
		fn getVolumeAndCOM obj =
		(
			local Volume= 0.0
			local Center= [0.0, 0.0, 0.0]
			local theMesh = snapshotasmesh obj
			local numFaces = theMesh.numfaces
			for i = 1 to numFaces do
			(
				local Face= getFace theMesh i
				local vert2 = getVert theMesh Face.z
				local vert1 = getVert theMesh Face.y
				local vert0 = getVert theMesh Face.x
				local dV = Dot (Cross (vert1 - vert0) (vert2 - vert0)) vert0
				Volume += dV
				Center += (vert0 + vert1 + vert2) * dV
			)
			delete theMesh
			Volume /= 6
			Center /= 24
			Center /= Volume
			return #(Volume,Center)
		)
		
		fn f_reverseArray &a_old =
		(
			local a_new = #()
			for i = a_old.count to 1 by -1 do
			(
				append a_new a_old[i]
			)
			a_old = a_new
		)
		
		fn f_sortArray a_objs doByName:false doByDistance:false doByVolume:false =
		(
			fn f_sortByName n1 n2 =
			(
				local result = n1.name < n2.name
				case of
				(
					(result): return -1
					(NOT result): return 1
					default: return 0
				)
			)
			
			fn f_sortByDistance n1 n2 =
			(
				local result = (length n1.pos)-(length n2.pos)
				case of
				(
					(result < 0.0): return -1
					(result > 0.0): return 1
					default: return 0
				)
			)
			
			--// Sort biggest to smallest by default (unless Reverse is toggled)
			fn f_sortByVolume n1 n2 =
			(
				local vol1 = (getVolumeAndCOM n1)[1]
				local vol2 = (getVolumeAndCOM n2)[1]
				local result = vol1 > vol2
				case of
				(
					(result): return -1
					(NOT result): return 1
					default: return 0
				)
			)
			if doByName do (qsort a_objs f_sortByName)
			if doByDistance do (qsort a_objs f_sortByDistance)
			if doByVolume do (qsort a_objs f_sortByVolume)
		)
		
		fn f_getPositionValues =
		(
			a_posX = #()
			a_posY = #()
			a_posZ = #()	
			for obj in a_selObjs do (append a_posX (obj.pos.x))
			for obj in a_selObjs do (append a_posY (obj.pos.y))
			for obj in a_selObjs do (append a_posZ (obj.pos.z))
		)
		
		fn f_getObjects =
		(
			a_selObjs = getcurrentselection()
			if NOT useSelOrder do
			(
				f_sortArray a_selObjs doByName:doByNameOrder doByDistance:doByDistanceOrder doByVolume:doByVolumeOrder
				if doReverseOrder do
				(
					f_reverseArray &a_selObjs
				)
			)
			if a_selObjs.count > 0 then
			(
				f_getPositionValues()
				return true
			)
			else
				return false
		)
		
		checkbox chkbx_useSelOrder "Use manual selection order" checked:false align:#left across:2 offset:[20,0] tooltip:"If checked, the order of your selection will be used otherwise the selection is sorted using a chosen method below"
		checkbox chkbx_reverseOrder "Reverse" checked:false align:#right offset:[-20,0] tooltip:"If checked, reverses the order of the selection based on the chosen Order-by method"
		group "Order by"
		(
			radiobuttons rdobtns_orderBy labels:#("Name","Distance","Volume") columns:3 align:#center offsets:#([-10,0],[0,0],[15,0]) tooltip:#("Re-orders the selection by object names","Re-orders the selection by distances from the origin","Re-orders the selection by volume size (biggest to smallest)")
		)
		group "Spread by Value"
		(
			radiobuttons rdobtns_ARspread labels:#("Absolute","Relative") offsets:#([-5,0],[5,0])
			spinner spn_spreadX "X: " range:[-1000.0,1000.0,0.0] align:#left width:100 across:3
			spinner spn_spreadY "Y: " range:[-1000.0,1000.0,0.0] align:#center width:100
			spinner spn_spreadZ "Z: " range:[-1000.0,1000.0,0.0] align:#right width:100
			button btn_applySpread ">>  Apply Spread To Selection  <<" height:25 width:300 align:#center
		)
		group "Offset by Value"
		(
			radiobuttons rdobtns_ARoffset labels:#("Absolute","Relative") default:2 offsets:#([-5,0],[5,0])
			label lbl_note "Note: for Absolute, values of zero are ignored"
			spinner spn_offsetX "X: " range:[-1000.0,1000.0,0.0] align:#left width:100 across:3
			spinner spn_offsetY "Y: " range:[-1000.0,1000.0,0.0] align:#center width:100
			spinner spn_offsetZ "Z: " range:[-1000.0,1000.0,0.0] align:#right width:100
			button btn_applyOffset ">>  Apply Offset To Selection  <<" height:25 width:300 align:#center
		)
		group "Spread Nudge"
		(
			spinner spn_spreadSensitivity "Spread sensitivity: " type:#float range:[0.01,10.0,0.1] scale:0.01 fieldwidth:45 align:#center
			slider sld_spreadX "Spread X:" range:[-100,100,0] type:#float align:#center ticks:2
			slider sld_spreadY "Spread Y:" range:[-100,100,0] type:#float align:#center ticks:2			
			slider sld_spreadZ "Spread Z:" range:[-100,100,0] type:#float align:#center ticks:2
			label lbl_tip1 "Right-Click slider to reset to zero (relative to first object)"
		)
		group "Offset Nudge"
		(
			spinner spn_offsetSensitivity "Offset sensitivity: " type:#float range:[0.01,10.0,0.1] scale:0.01 fieldwidth:45 align:#center
			slider sld_offsetX "Offset X:" range:[-100,100,0] type:#float align:#center ticks:2
			slider sld_offsetY "Offset Y:" range:[-100,100,0] type:#float align:#center ticks:2			
			slider sld_offsetZ "Offset Z:" range:[-100,100,0] type:#float align:#center ticks:2
			label lbl_tip2 "Right-Click slider to align to origin"
		)
		button btn_about "About" align:#center width:40 height:20
		
		on spreadObjects_Rollout open do
		(
			f_getObjects()
		)
		
		on spreadObjects_Rollout moved pos do
		(
			g_LBTools_spreadObjs.dPos = pos
			setIniSetting g_LBTools_spreadObjs.spreadObjectsINI "Spread Objects" "Position" (pos as string)
		)
		
		on spreadObjects_Rollout close do
		(
			try(destroyDialog g_LBTools_spreadObjs.aboutScript_rollout)catch()
		)
		
		on btn_about pressed do
		(
			try(destroyDialog g_LBTools_spreadObjs.aboutScript_rollout)catch()
			createDialog g_LBTools_spreadObjs.aboutScript_rollout
		)
		
		on chkbx_useSelOrder changed state do
		(
			if state then
			(
				useSelOrder = true
				rdobtns_orderBy.enabled = false
			)
			else
			(
				useSelOrder = false
				rdobtns_orderBy.enabled = true
			)
		)
		
		on rdobtns_orderBy changed state do
		(
			if state == 1 do
			(
				doByNameOrder = true
				doByDistanceOrder = false
				doByVolumeOrder = false
			)
			if state == 2 do
			(
				doByNameOrder = false
				doByDistanceOrder = true
				doByVolumeOrder = false
			)
			if state == 3 do
			(
				doByNameOrder = false
				doByDistanceOrder = false
				doByVolumeOrder = true
			)
		)
		
		on chkbx_reverseOrder changed state do
		(
			if state then
				doReverseOrder = true
			else
				doReverseOrder = false
		)
		
		on spn_spreadX changed val do (spreadX_val = val)
		on spn_spreadY changed val do (spreadY_val = val)
		on spn_spreadZ changed val do (spreadZ_val = val)

		on btn_applySpread pressed do
		(
			if (f_getObjects()) do
			(
				undo "SpreadObjects" on
				(
					if (f_getObjects()) do
					(
						--// If absolute, align all the objects to the first one and reset the values
						if rdobtns_ARspread.state == 1 do
						(
							for i = 2 to a_selObjs.count do
							(
								a_selObjs[i].position.x = a_selObjs[1].position.x
								a_selObjs[i].position.y = a_selObjs[1].position.y
								a_selObjs[i].position.z = a_selObjs[1].position.z
							)
							f_getPositionValues()
						)
						for i = 2 to a_selObjs.count do
						(
							a_selObjs[i].pos.x = a_posX[i] + (spreadX_val * (i-1))
							a_selObjs[i].pos.y = a_posY[i] + (spreadY_val * (i-1))
							a_selObjs[i].pos.z = a_posZ[i] + (spreadZ_val * (i-1))
						)
					)
				)
			)
		)
		
		on spn_offsetX changed val do (offsetX_val = val)
		on spn_offsetY changed val do (offsetY_val = val)
		on spn_offsetZ changed val do (offsetZ_val = val)
		
		on btn_applyOffset pressed do
		(
			undo "OffsetObjects" on
			(
				if (f_getObjects()) do
				(
					if rdobtns_ARoffset.state == 1 then --// do absolute
					(
						for obj in a_selObjs do
						(
							if (offsetX_val != 0) do (obj.pos.x = offsetX_val)
							if (offsetY_val != 0) do (obj.pos.y = offsetY_val)
							if (offsetZ_val != 0) do (obj.pos.z = offsetZ_val)
						)
					)
					else --// do relative
					(
						for obj in a_selObjs do
						(
							if (offsetX_val != 0) do (obj.pos.x += offsetX_val)
							if (offsetY_val != 0) do (obj.pos.y += offsetY_val)
							if (offsetZ_val != 0) do (obj.pos.z += offsetZ_val)
						)
					)
				)
			)
		)	
	
		/************************** X-AXIS ***************************/
		
		on sld_spreadX changed val do
		(
			if (f_getObjects()) do
			(
				undo "SpreadX" on
				(
					for i = 2 to a_selObjs.count do
					(
						a_selObjs[i].pos.x = a_posX[i] + (val * ((i-1) * spn_spreadSensitivity.value))
					)
				)
			)
		)
		
		on sld_spreadX rightclick do
		(
			if (f_getObjects()) do
			(
				for i = 2 to a_selObjs.count do
				(
					a_selObjs[i].position.x = a_selObjs[1].position.x
				)
			)
		)
		
		on sld_spreadX buttonup do
		(
			sld_spreadX.value = 0
			f_getPositionValues()
		)
		
		on sld_offsetX changed val do
		(
			if (f_getObjects()) do
			(
				undo "OffsetX" on
				(
					for obj in a_selObjs do
					(
						obj.pos.x += (val * spn_offsetSensitivity.value)
					)
				)
			)
		)
		
		on sld_offsetX buttonup do
		(
			sld_offsetX.value = 0
		)
		
		on sld_offsetX rightclick do
		(
			if (f_getObjects()) do
			(
				for obj in a_selObjs do
				(
					obj.position.x = 0
				)
			)
		)
		
		/************************** Y-AXIS ***************************/
		
		on sld_spreadY changed val do
		(
			if (f_getObjects()) do
			(
				undo "SpreadY" on
				(
					for i = 2 to a_selObjs.count do
					(
						a_selObjs[i].pos.y = a_posY[i] + (val * ((i-1) * spn_spreadSensitivity.value))
					)
				)
			)
		)
		
		on sld_spreadY rightclick do
		(
			if (f_getObjects()) do
			(
				for i = 2 to a_selObjs.count do
				(
					a_selObjs[i].position.y = a_selObjs[1].position.y
				)
			)
		)
		
		on sld_spreadY buttonup do
		(
			sld_spreadY.value = 0
			f_getPositionValues()
		)
		
		on sld_offsetY changed val do
		(
			if (f_getObjects()) do
			(
				undo "OffsetY" on
				(
					for obj in a_selObjs do
					(
						obj.pos.y += (val * spn_offsetSensitivity.value)
					)
				)
			)
		)
		
		on sld_offsetY buttonup do
		(
			sld_offsetY.value = 0
		)
		
		on sld_offsetY rightclick do
		(
			if (f_getObjects()) do
			(
				for obj in a_selObjs do
				(
					obj.position.y = 0
				)
			)
		)
		
		/************************** Z-AXIS ***************************/
		
		on sld_spreadZ changed val do
		(
			if (f_getObjects()) do
			(
				undo "SpreadZ" on
				(
					for i = 2 to a_selObjs.count do
					(
						a_selObjs[i].pos.z = a_posZ[i] + (val * ((i-1) * spn_spreadSensitivity.value))
					)
				)
			)
		)
		
		on sld_spreadZ rightclick do
		(
			if (f_getObjects()) do
			(
				for i = 2 to a_selObjs.count do
				(
					a_selObjs[i].position.z = a_selObjs[1].position.z
				)
			)
		)
		
		on sld_spreadZ buttonup do
		(
			sld_spreadZ.value = 0
			f_getPositionValues()
		)
		
		on sld_offsetZ changed val do
		(
			if (f_getObjects()) do
			(
				undo "OffsetZ" on
				(
					for obj in a_selObjs do
					(
						obj.pos.z += (val * spn_offsetSensitivity.value)
					)
				)
			)
		)
		
		on sld_offsetZ buttonup do
		(
			sld_offsetZ.value = 0
		)
		
		on sld_offsetZ rightclick do
		(
			if (f_getObjects()) do
			(
				for obj in a_selObjs do
				(
					obj.position.z = 0
				)
			)
		)
	) --// end rollout
	
	rollout aboutScript_rollout "About"
	(
		label lbl_author "Author - Lorne Brooks"
		hyperLink myEmail "lbtools2k@gmail.com" color:(color 0 100 255) hovercolor:(color 0 150 255) visitedcolor:(color 0 150 255) align:#center address:"mailto:lbtools2k@gmail.com"
		label lbl_version "Version 1.1 - 15 Nov 2018"
	) --// end rollout
	
	fn f_initializeINI =
	(
		if NOT (doesFileExist g_LBTools_spreadObjs.spreadObjectsINI) then
		(
			setIniSetting g_LBTools_spreadObjs.spreadObjectsINI "Spread Objects" "Position" ((mouse.screenpos + [-75,-15]) as string)
			g_LBTools_spreadObjs.dPos = mouse.screenpos + [-75,-15]
		)
		else
		(
			g_LBTools_spreadObjs.dPos = execute(getIniSetting g_LBTools_spreadObjs.spreadObjectsINI "Spread Objects" "Position")
		)
	)
	
	--// Load UI
	fn f_openSpreadObjectsUI =
	(
		try(destroyDialog spreadObjects_Rollout)catch()
		createDialog spreadObjects_Rollout width:320 height:765 pos:g_LBTools_spreadObjs.dPos
	)
	
	f_initializeINI()
	f_openSpreadObjectsUI()
)

) --// end struct

g_LBTools_spreadObjs = s_LBTools_spreadObjs()

/*********************************************************************************************************************************************************************/

macroscript spreadObjects
	category:"LB Tools"
	buttontext:"{ SO }"
	icon:#("LBTools_SpreadObjects",1)
	tooltip:"Spread Objects 1.1"
(
	g_LBTools_spreadObjs.f_spreadObjects()
)

/*********************************************************************************************************************************************************************/